

# TASK EXCEPTIONS
class ProductNotProducedException(Exception):
    pass


class ProductNotUpdatedException(Exception):
    pass


class DependencyNotMetException(Exception):
    pass


class TaskRaisedExceptionException(Exception):
    pass

class UnknownStatusException(Exception):
    pass


# TASKHANDLER EXCEPTION
class TaskNotInQueueException(Exception):
    pass

class ProductAlreadyRegisteredException(Exception):
    pass

class DependencyNotAvailableException(Exception):
    pass




import concurrent.futures
from abc import ABC, abstractmethod
from concurrent.futures.thread import ThreadPoolExecutor
from typing import List
import submitit
from attrs import frozen

from .Task import Task


class AbstractTaskExecutor(ABC):
    @abstractmethod
    def submit(self, task, task_dependencies:List[Task]=None):
        ...

    @abstractmethod
    def wait_for_all(self):
        ...

    def handles_dependencies(self):
        ...

@frozen
class DemoTaskExecutor(AbstractTaskExecutor):
    def submit(self, task, task_dependencies:List[Task]=None):
        task.run()

    def wait_for_all(self):
        pass

    def handles_dependencies(self):
        return False


class ParallelExecutor(AbstractTaskExecutor):

    def __init__(self, internal_executor:concurrent.futures.Executor=None, **kwargs):
        self.internal_executor = internal_executor if internal_executor is not None else ThreadPoolExecutor()
        self.running_jobs = []
        self.running_tasks = []
        print("depio-ParallelExecutor initialized")


    def submit(self, task, task_dependencies:List[Task]=None):
        job = self.internal_executor.submit(task.run)
        self.running_jobs.append(job)
        self.running_tasks.append(task)
        return

    def wait_for_all(self):
        for job in self.running_jobs:
            job.result()

    def handles_dependencies(self):
        return False


TIME_IN_MINUTES = 60 * 48  # 48 hours in minutes
DEFAULT_PARAMS = {
    "slurm_time": TIME_IN_MINUTES,
    "slurm_partition": "gpu",
    "slurm_mem": 32,
    "gpus_per_node": 0
}

class SubmitItExecutor(AbstractTaskExecutor):

    def __init__(self, internal_executor=None, **kwargs):

        self.internal_executor = internal_executor if internal_executor is not None else submitit.AutoExecutor()
        self.internal_executor.update_parameters(**DEFAULT_PARAMS)
        self.running_jobs = []
        self.running_tasks = []
        print("depio-SubmitItExecutor initialized")


    def submit(self, task, task_dependencies:List[Task]=None):
        slurm_additional_parameters = {}
        afterok:List[str] = [f"{t.slurmjob.job_id}" for t in task_dependencies]
        if len(afterok) > 0:
            slurm_additional_parameters["dependency"] = f"afterok:{':'.join(afterok)}"

        self.internal_executor.update_parameters(**DEFAULT_PARAMS, slurm_additional_parameters=slurm_additional_parameters)
        job = self.internal_executor.submit(task.run)
        task.slurmjob = job
        self.running_jobs.append(job)
        self.running_tasks.append(task)
        return

    def wait_for_all(self):
        for job in self.running_jobs:
            job.result()

    def handles_dependencies(self):
        return True


__all__ = [AbstractTaskExecutor, ParallelExecutor, DemoTaskExecutor, SubmitItExecutor]



from __future__ import annotations

import pathlib
import typing
from io import StringIO
from os.path import getmtime
from typing import List, Dict, Callable, get_origin, Annotated, get_args, Union
import sys

from nltk.corpus import dependency_treebank

from .TaskStatus import TaskStatus, TERMINAL_STATES, SUCCESSFUL_TERMINAL_STATES, FAILED_TERMINAL_STATES
from .stdio_helpers import redirect, stop_redirect
from .exceptions import ProductNotProducedException, TaskRaisedExceptionException, UnknownStatusException, ProductNotUpdatedException, \
    DependencyNotMetException


class Product():
    pass


class Dependency():
    pass


status_colors = {
    TaskStatus.WAITING: 'blue',
    TaskStatus.DEPFAILED: 'red',
    TaskStatus.PENDING: 'blue',
    TaskStatus.RUNNING: 'yellow',
    TaskStatus.FINISHED: 'green',
    TaskStatus.SKIPPED: 'green',
    TaskStatus.HOLD: 'white',
    TaskStatus.FAILED: 'red',
    TaskStatus.CANCELED: 'white',
    TaskStatus.UNKNOWN: 'white'
}


def python_version_is_greater_equal_3_10():
    return sys.version_info.major > 3 and sys.version_info.minor >= 10


# from https://stackoverflow.com/questions/218616/how-to-get-method-parameter-names
def _get_args_dict(fn, args, kwargs):
    args_names = fn.__code__.co_varnames[:fn.__code__.co_argcount]
    return {**dict(zip(args_names, args)), **kwargs}


def _parse_annotation_for_metaclass(func, metaclass):
    if python_version_is_greater_equal_3_10():
        # For python 3.10 and newer
        # annotations = inspect.get_annotations(func)

        # According to https://docs.python.org/3/howto/annotations.html this is best practice now.
        annotations = getattr(func, '__annotations__', None)
    else:
        # For python 3.9 and older
        if isinstance(func, type):
            annotations = func.__dict__.get('__annotations__', None)
        else:
            annotations = getattr(func, '__annotations__', None)

    results = []

    for name, annotation in annotations.items():
        if get_origin(annotation) is Annotated:
            args = get_args(annotation)
            if len(args) <= 1:
                continue

            metadata = args[1:]
            if any(meta is metaclass for meta in metadata):
                results.append(name)

    return results


def _get_not_updated_products(product_timestamps_after_running: typing.Dict, product_timestamps_before_running: typing.Dict) -> typing.List[str]:
    # Calculate the not updated products
    not_updated_products = []
    for product, before_timestamp in product_timestamps_before_running.items():
        after_timestamp = product_timestamps_after_running.get(product)
        if before_timestamp == after_timestamp:
            not_updated_products.append(product)

    return not_updated_products


class Task:
    def __init__(self, name: str, func: Callable, func_args: List = None, func_kwargs: List = None,
                 produces: List[pathlib.Path] = None, depends_on: List[Union[pathlib.Path, Task]] = None):
        produces = produces if produces is not None else []
        depends_on = depends_on if depends_on is not None else []

        self._status: TaskStatus = TaskStatus.WAITING
        self.name: str = name
        self.queue_id: int = None
        self.slurmjob = None
        self.func = func
        self.func_args: List = func_args or []
        self.func_kwargs: Dict = func_kwargs or {}

        # Parse dependencies and products from the annotations and merge with args
        self.products_args = _parse_annotation_for_metaclass(func, Product)
        self.dependencies_args = _parse_annotation_for_metaclass(func, Dependency)
        args_dict = _get_args_dict(func, self.func_args, self.func_kwargs)
        self.products: List[pathlib.Path] = [args_dict[argname] for argname in self.products_args if argname in args_dict] + produces
        self.dependencies: List[Union[Task, pathlib.Path]] = [args_dict[argname] for argname in self.dependencies_args if argname in args_dict] + depends_on

        self.stdout = StringIO()
        self.slurmjob = None
        self._slurmid = None

    def __str__(self):
        return f"Task:{self.name}"

    def run(self):

        redirect(self.stdout)

        # Check if all path dependencies are met
        not_existing_path_dependencies: List[str] = [str(dependency) for dependency in self.path_dependencies if not dependency.exists()]
        if len(not_existing_path_dependencies) > 0:
            self._status = TaskStatus.FAILED
            raise DependencyNotMetException(f"Task {self.name}: Dependency/ies {not_existing_path_dependencies} not met.")

        # Store the last-modification timestamp of the already existing products.
        product_timestamps_before_running: Dict[str, float] = {str(product): getmtime(product) for product in self.products if product.exists()}

        # Call the actual function
        self._status = TaskStatus.RUNNING

        try:
            self.func(*self.func_args, **self.func_kwargs)
        except Exception as e:
            self._status = TaskStatus.FAILED
            raise TaskRaisedExceptionException(e)
        finally:
            stop_redirect()

        # Check if any product does not exist.
        not_existing_products: List[str] = [str(product) for product in self.products if not product.exists()]
        if len(not_existing_products) > 0:
            self._status = TaskStatus.FAILED
            raise ProductNotProducedException(f"Task {self.name}: Product/s {not_existing_products} not produced.")

        # Check if any product has not been updated.
        product_timestamps_after_running: Dict = {str(product): getmtime(product) for product in self.products if product.exists()}

        not_updated_products = _get_not_updated_products(product_timestamps_after_running, product_timestamps_before_running)

        if len(not_updated_products) > 0:
            self._status = TaskStatus.FAILED
            raise ProductNotUpdatedException(f"Task {self.name}: Product/s {not_updated_products} not updated.")

        self._status = TaskStatus.FINISHED

    def _update_by_slurmjob(self):
        assert self.slurmjob is not None

        self.slurmjob.watcher.update()
        info = self.slurmjob.get_info()

        self._slurmid = f"{int(self.slurmjob.job_id):d}-{int(self.slurmjob.task_id):d}"

        if self.slurmjob.state in ['RUNNING', 'CONFIGURING', 'COMPLETING', 'STAGE_OUT']:
            self._status = TaskStatus.RUNNING
        elif self.slurmjob.state in ['FAILED', 'BOOT_FAIL', 'DEADLINE', 'NODE_FAIL', 'OUT_OF_MEMORY', 'PREEMPTED', 'SPECIAL_EXIT', 'STOPPED',
                                     'SUSPENDED', 'TIMEOUT']:
            self._status = TaskStatus.FAILED
        elif self.slurmjob.state in ['READY', 'PENDING', 'REQUEUE_FED', 'REQUEUED']:
            self._status = TaskStatus.PENDING
        elif self.slurmjob.state == 'CANCELED':
            self._status = TaskStatus.CANCELED
        elif self.slurmjob.state in ['COMPLETED']:
            self._status = TaskStatus.FINISHED
        elif self.slurmjob.state in ['RESV_DEL_HOLD', 'REQUEUE_HOLD', 'RESIZING', 'REVOKED', 'SIGNALING']:
            self._status = TaskStatus.HOLD
        elif self.slurmjob.state == 'UNKNOWN':
            self._status = TaskStatus.UNKNOWN
        else:
            raise Exception(f"Unknown slurmjob status! slurmjob.done {self.slurmjob.done}, slurmjob.state {self.slurmjob.state} ")

    @property
    def slurmjob_status(self):
        if not self.slurmjob is None:
            return self.slurmjob.state
        else:
            return ""

    def statuscolor(self, s: TaskStatus = None) -> str:
        if s is None: s = self._status
        if s in status_colors:
            return status_colors[s]
        else:
            raise UnknownStatusException("Status {} is unknown.".format(s))

    def statustext(self, s: TaskStatus = None) -> str:
        if s is None: s = self._status
        status_messages = {
            TaskStatus.WAITING: lambda: 'waiting' + (f" for {[d.queue_id for d in self.task_dependencies if not d.is_in_terminal_state]}" if len(
                [d for d in self.task_dependencies if not d.is_in_terminal_state]) > 1 else ""),
            TaskStatus.DEPFAILED: lambda: 'dep. failed' + (
                f" at {[d.queue_id for d in self.task_dependencies if d.is_in_failed_terminal_state]}" if len(
                    [d for d in self.task_dependencies if d.is_in_failed_terminal_state]) > 1 else ""),
            TaskStatus.PENDING: lambda: 'pending',
            TaskStatus.RUNNING: lambda: 'running',
            TaskStatus.FINISHED: lambda: 'finished',
            TaskStatus.SKIPPED: lambda: 'skipped',
            TaskStatus.HOLD: lambda: 'hold',
            TaskStatus.FAILED: lambda: 'failed',
            TaskStatus.CANCELED: lambda: 'cancelled',
            TaskStatus.UNKNOWN: lambda: 'unknown'
        }
        try:
            return status_messages[s]()
        except KeyError:
            raise UnknownStatusException(f"Status {s} is unknown.")

    @property
    def status(self):
        s = self._status  # Fix status as temporary to return a consistent tuple
        return s, self.statustext(s), self.statuscolor(s)

    @property
    def is_in_terminal_state(self) -> bool:
        return self._status in TERMINAL_STATES

    @property
    def is_in_successful_terminal_state(self) -> bool:
        return self._status in SUCCESSFUL_TERMINAL_STATES

    @property
    def is_in_failed_terminal_state(self) -> bool:
        return self._status in FAILED_TERMINAL_STATES

    def set_to_depfailed(self) -> None:
        self._status = TaskStatus.DEPFAILED

    @property
    def id(self) -> str:
        return f"{self.queue_id: 4d}"

    @property
    def slurmid(self) -> str:
        if not self.slurmjob is None:
            self._update_by_slurmjob()
            return f"{self._slurmid}"
        else:
            return ""

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            b = (self.func == other.func
                 and self.func_args == other.func_args
                 and self.func_kwargs == other.func_kwargs
                 and self.name == other.name)
            return b
        else:
            return False


__all__ = [Task, Product, Dependency, _get_not_updated_products]



from .Task import Task
from .Pipeline import Pipeline




def task(name : str, *dec_args, **dec_kwargs):
    def wrapper(func):
        def decorator(*func_args, **func_kwargs):
            # Create and add the task
            t = Task(name, func=func, func_args=func_args, func_kwargs=func_kwargs)

            # Not call the function

            return t
        return decorator
    return wrapper

__all__ = [task]


from typing import Set, Dict, List
from pathlib import Path
import time
import sys
from termcolor import colored
from tabulate import tabulate

from .stdio_helpers import enable_proxy
from .Task import Task
from .TaskStatus import TaskStatus
from .Executors import AbstractTaskExecutor
from .exceptions import ProductAlreadyRegisteredException, TaskNotInQueueException, DependencyNotAvailableException


class Pipeline:
    def __init__(self, depioExecutor: AbstractTaskExecutor, name: str = "NONAME",
                 clear_screen: bool = True,
                 quiet: bool = False,
                 refreshrate: float = 1.0):
        self.CLEAR_SCREEN: bool = clear_screen
        self.QUIET: bool = quiet
        self.REFRESHRATE: float = refreshrate

        self.name: str = name
        self.submitted_tasks: List[Task] = None
        self.tasks: List[Task] = []
        self.depioExecutor: AbstractTaskExecutor = depioExecutor
        self.registered_products: Set[Path] = set()
        if not self.QUIET: print("Pipeline initialized")

    def add_task(self, task: Task) -> None:
        # Check if the exact task is already registered
        for registered_task in self.tasks:
            if task == registered_task:
                return registered_task


        # Check is a product is already registered
        products_already_registered: List[str] = [str(p) for p in task.products if str(p) in set(map(str, self.registered_products))]
        if len(products_already_registered) > 0:
            raise ProductAlreadyRegisteredException(
                f"The product\s {products_already_registered} is/are already registered. Each output can only be registered from one task.")

        # Check if the task dependencies are registered already
        missing_tasks: List[Task] = [t for t in task.dependencies if isinstance(t, Task) and t not in self.tasks]
        if len(missing_tasks) > 0:
            raise TaskNotInQueueException("Add the tasks into the queue in the correct order. The following task/s is/are missing: {missing_tasks}.")

        # Register products
        self.registered_products.update(task.products)

        # Register task
        self.tasks.append(task)
        task.queue_id = len(self.tasks)
        return task

    def _solve_order(self) -> None:
        # Generate a task to product mapping.
        product_to_task: Dict[Path, Task] = {product: task for task in self.tasks for product in task.products}

        # Add the dependencies to the tasks
        for task in self.tasks:
            # Verify that each dependency is available and add if yes.
            unavailable_dependencies = [d for d in task.dependencies
                                        if d not in self.tasks and d not in product_to_task and not d.exists()]
            if len(unavailable_dependencies) > 0:
                raise DependencyNotAvailableException(f"Dependency/ies '{unavailable_dependencies}' do/es not exist and can not be produced.")

            # Add the tasks that produce path_deps and remove such deps from the path_deps
            task.task_dependencies = ([product_to_task[d] for d in task.dependencies if isinstance(d, Path) and d in product_to_task]
                                      + [d for d in task.dependencies if isinstance(d,Task)])
            task.path_dependencies = [d for d in task.dependencies if isinstance(d, Path) and d not in product_to_task]

    def _submit_task(self, task: Task) -> bool:
        """
        Submits the task to the extractor if all dependencies are available.
        Otherwise, the function is called recursively for each dependency.

        :param task:
        :return:
        """
        if task in self.submitted_tasks: return

        all_dependencies_are_available = True
        is_new_depfail_found = False


        # Execute and check all dependencies first
        for t_dep in task.task_dependencies:
            assert isinstance(t_dep, Task)
            self._submit_task(t_dep)  # Recursive call for dependency
            if not t_dep.is_in_successful_terminal_state:
                all_dependencies_are_available = False

            if t_dep.is_in_failed_terminal_state and not task.is_in_failed_terminal_state:
                # If the task is not already in failed state:
                task.set_to_depfailed()  # set to depfailed
                is_new_depfail_found = True  # Remember that we propagated dependency failures

        for p_dep in task.path_dependencies:
            assert isinstance(p_dep, Path)
            if not p_dep.exists():
                all_dependencies_are_available = False

                if not task.is_in_failed_terminal_state:
                    # If the task is not already in failed state:
                    task.set_to_depfailed()  # set to depfailed
                    is_new_depfail_found = True  # Remember that we propagated dependency failures

        # Execute the task if all dependencies are given
        if all_dependencies_are_available or self.depioExecutor.handles_dependencies():
            self.depioExecutor.submit(task, task.task_dependencies)
            self.submitted_tasks.append(task)

        return is_new_depfail_found

    def run(self) -> None:
        enable_proxy()

        self._solve_order()
        self.submitted_tasks: List[Task] = []

        while True:
            try:
                # Iterate over all tasks in the queue until now new depfail is found
                while True:
                    if all(not self._submit_task(task) for task in self.tasks):
                        break

                # Check the status of all tasks
                all_tasks_in_terminal_state = all(task.is_in_terminal_state for task in self.tasks)
                if not self.QUIET: self._print_tasks()
                if all_tasks_in_terminal_state:
                    if any(task.is_in_failed_terminal_state for task in self.tasks):
                        self.exit_with_failed_tasks()
                    else:
                        self.exit_successful()

            except KeyboardInterrupt:
                print("Stopping execution bc of keyboard interrupt!")
                exit(1)
            time.sleep(self.REFRESHRATE)

    def _get_text_for_task(self, task):
        status = task.status

        formatted_status = colored(f"{status[1].upper():<{len('DEP. FAILED')}s}", status[2])
        formatted_slurmstatus = colored(f"{task.slurmjob_status:<{len('OUT_OF_MEMORY')}s}", status[2])
        return [
            task.id,
            task.name,
            task.slurmid,
            formatted_slurmstatus,
            formatted_status,
            [t.queue_id for t in task.task_dependencies],
            [str(d) for d in task.dependencies if isinstance(d, Path)],
            [str(p) for p in task.products]
        ]

    def _clear_screen(self):
        if self.CLEAR_SCREEN: sys.stdout.write("\033[2J\033[H")

    def _print_tasks(self):
        self._clear_screen()
        headers = ["ID", "Name", "Slurm ID", "Slurm Status", "Status", "Task Deps.", "Path Deps.", "Products"]
        tasks_data = [self._get_text_for_task(task) for task in self.tasks]
        table_str = tabulate(tasks_data, headers=headers, tablefmt="plain")
        print()
        print("Tasks:")
        print(table_str)

    def exit_with_failed_tasks(self) -> None:
        print()
        failed_tasks = [
            [task.id, task.name, task.slurmid, task.status[1]]
            for task in self.tasks if task.status[0] == TaskStatus.FAILED
        ]

        if failed_tasks:
            headers = ["Task ID", "Name", "Slurm ID", "Status"]
            print("---> Summary of Failed Tasks:")
            print()

            for task in self.tasks:
                if task.status[0] == TaskStatus.FAILED:
                    print(f"Details for Task ID: {task.id} - Name: {task.name}")
                    print(tabulate([[task.stdout.getvalue()]], headers=["STDOUT"], tablefmt="grid"))

        print("Exit.")
        exit(1)

    def exit_successful(self) -> None:
        print("All jobs done! Exit.")
        exit(0)


__all__ = [Pipeline]
